#!/usr/bin/env python3

from configparser import ConfigParser
import multiprocessing
import threading
from threading import current_thread
import datetime
import errno
import getopt
import json
import logging
import math
import os
import pwd
import sqlite3
import sys
import time

#config file full path
config_path = '/usr/local/safesquid/api/config.ini';

# set config initialization
config = ConfigParser()

# parse existing file
config.read(config_path)

# read values from a section
log_path = config.get('log_info', 'log_path')
log_file = config.get('log_info', 'log_file')

report_dir = config.get('report_path', 'report_dir')
db_resource = config.get('database', 'resource_db')
db_src_dir = config.get('database', 'src_path')
main_db_str = config.get('database_structure', 'main_db')
attach_db_str = config.get('database_structure', 'attach_db')

user = config.get('permission', 'user')
group = config.get('permission', 'group')

# logs to be written to file
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# create a file handler
log_file_path = log_path + log_file
handler = logging.FileHandler(log_file_path)
handler.setLevel(logging.INFO)

# create a logging format
formatter = logging.Formatter('%(asctime)s - %(name)s - [ %(funcName)s():%(lineno)3s ] - %(levelname)s - %(message)s')
handler.setFormatter(formatter)

# add the handlers to the logger
logger.addHandler(handler)

from_date = ''
to_date = ''
total_days = ''
date_format = "%d.%m.%Y %H:%M:%S"
days_path = []

def epoch_time(date_time=None):
    if date_time is None:
        date_time = time.strftime("%d.%m.%Y %H:%M:%S")

    pattern = '%d.%m.%Y %H:%M:%S'
    epoch = int(time.mktime(time.strptime(date_time, date_format)))
    return epoch

#get period
def getdays(fromdata, todate):    
    a = datetime.datetime.strptime(fromdata, date_format)
    b = datetime.datetime.strptime(todate, date_format)

    delta = b - a
    return delta.days

def write_file(file_name, data):
    f = open(file_name, 'w')
    f.write(data)
    f.close()

#get size
def convert_size(size_bytes):
    if not size_bytes:
        return '0B'

    if (size_bytes == 0):
        return '0B'

    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return '%s %s' % (s, size_name[i])

def getSize(filename):
    try:
        st = os.stat(filename)
        return convert_size(st.st_size)
    except (OSError, ValueError) as e:
        logger.error(str(e))
        print (filename + ' filename is not available')
        sys.exit(2)

def check_directory(path):
    if not os.path.exists(path):
        try:
            uid, gid = pwd.getpwnam(user).pw_uid, pwd.getpwnam(group).pw_uid
            logger.info('creating directory %s with [%s:%s]', path, user, group)
            os.makedirs(path)
            os.chown(path, uid, gid)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                logger.error('creating directory %s failed', path)
                raise

#file to attach count
def file_to_attach(date_from_get,date_to_get):    
    
    file_attach = []
    
    ago = datetime.datetime.strptime(date_from_get, date_format)
    ago1 = datetime.datetime.strptime(date_to_get, date_format)
    
    for root, dirs,files in os.walk(db_src_dir):
        for fname in files:
            path = os.path.join(root, fname)
            try :
                st = os.stat(path)
                mtime = datetime.datetime.fromtimestamp(st.st_mtime)
                filename, file_extension = os.path.splitext(path)

                if mtime > ago and mtime < ago1:                
                    if file_extension != ".db-wal" and file_extension != ".db-shm" and not filename.endswith("safesquid"): 
                        file_attach.append(path)            
                        #logger.info('Generating report from %s [%s]', path, getSize(path))
            except FileNotFoundError as err:            
                pass
    
    logger.info("[%d] on %s ",len(file_attach),ago)
    return file_attach
    
#fetch command line argument so that report can be generated.
def arguments():
    global from_date, to_date, total_days
    def usage():        
        print ('Usage : ./app_handler.py -d <date>                                      -- Generate for particular date')
        print ('        ./app_handler.py -s <start date> -e <end date>                  -- Generate between time range')
        print ('Date formatte will be in dd.mm.yyyy Eg : 03.08.1990')
        print ('If you are using time range. Try Generate reports in packs of 7 days')

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hs:e:d:")
    except getopt.GetoptError as err:
        print (str(err))
        usage()
        sys.exit(2)

    if (len(opts) == 0):
        logger.error('No arg passed')
        usage()
        sys.exit(2)
    else:
        logger.info('total arg passed: %s', len(opts))

        for opt, arg in opts:
            if opt == '-h':
                usage()
                sys.exit(2)
            elif opt in ("-s"):
                from_date = time.strftime(arg + " 00:00:00")                
            elif opt in ("-e"):
                to_date = time.strftime(arg + " 23:59:59")                
            elif opt in ("-d"):
                from_date = time.strftime(arg + " 00:00:00")
                to_date = time.strftime(arg + " 23:59:59")
        
        if not from_date:
            lastHourDateTime = datetime.datetime.now() - datetime.timedelta(hours=1)
            from_date = lastHourDateTime.strftime("%d.%m.%Y %H:00:00")
            del lastHourDateTime

        if not to_date:
            to_date = time.strftime("%d.%m.%Y %H:00:00")
        
        total_days = getdays(from_date, to_date)        
        
        if total_days < 0:
            logger.error('end date defination is wrong')
            print ('Incorrect end date . Start date [' + from_date + '] is greater than end date [' + to_date + ']')
            sys.exit(2)
        else:
            total_days = total_days  + 1
        
        logger.info("start date[%s], End date[%s] with days to generate %i", from_date, to_date, total_days)

def delete_file(path):
    try:
        if os.path.isfile(path):
            os.remove(path) 
        elif os.path.isdir(path):
            shutil.rmtree(path) 
    except OSError as e: 
        print ("Error: %s - %s." % (e.filename, e.strerror))

def reverse_lookup(array_name,value):
    try:
        return list(array_name.keys())[list(array_name.values()).index(value)]
    except ValueError:
        return None        
        
def merge_db(date_cnt):
    global days_path
    start_generation_m = time.time()
    lastHourDateTime = datetime.datetime.strptime(to_date, date_format) - datetime.timedelta(days=date_cnt)
    from_date_ = lastHourDateTime.strftime("%d.%m.%Y 00:00:00")
    to_date_ = lastHourDateTime.strftime("%d.%m.%Y 23:59:59")
        
    db_name = lastHourDateTime.strftime("%d-%m-%Y")
        
    file_attach = file_to_attach(from_date_,to_date_)        
    new_db = db_name + 'safesquid.db'        
    merge_full_path = db_src_dir +'/'+ new_db
    
    report_str = db_name.split('-')

    report_path = report_dir + '/' + report_str[2] + '/' + report_str[1] + '/' + report_str[0]
    check_directory(report_path)
    days_path.append(report_path)
        
    try:
        os.remove(merge_full_path)
    except OSError:
        pass
        
    conn=sqlite3.connect(merge_full_path)        
        
    try:
        conn.row_factory = sqlite3.Row
        curs = conn.cursor()                     
        curs.executescript(main_db_str[1:-1])
        recs = curs.fetchall()
    except sqlite3.OperationalError as er:
        logger.error('[%s] %s',db_name, er)
    except sqlite3.Error as er:
        logger.error('[%s] %s',db_name, er)
        
    try:
        m0 = file_attach.index("/var/db/safesquid/report/safesquid2.db")
        del file_attach[m0]
    except ValueError:            
        continue_flg = 1
        
    for i in range(len(file_attach)):
        
        conn.row_factory = sqlite3.Row
        curs = conn.cursor() 
        
        try:                                
            curs.executescript("attach database '"+file_attach[i]+"' as merge;"+attach_db_str[1:-1])        
        except sqlite3.OperationalError as er:
            logger.error('[%s:%s] %s',db_name,file_attach[i], er)
            curs.executescript("detach merge;")        
        except sqlite3.Error as er:
            logger.error('[%s:%s] %s',db_name,file_attach[i], er)
            curs.executescript("detach merge;")        
        recs = curs.fetchall()
        
    conn.close()    
    
    end_generation_m = time.time()
    exhicution_time_m = end_generation_m - start_generation_m
    logger.info("Exhicution time taken for %s is %s",db_name,exhicution_time_m)

def query_builder(date_cnt):
    lastHourDateTime = datetime.datetime.strptime(to_date, date_format) - datetime.timedelta(days=date_cnt)
    from_date_ = lastHourDateTime.strftime("%d.%m.%Y 00:00:00")
    to_date_ = lastHourDateTime.strftime("%d.%m.%Y 23:59:59")
    
    db_name = lastHourDateTime.strftime("%d-%m-%Y")
    
    new_db = db_name + 'safesquid.db'        
    merge_full_path = db_src_dir +'/'+ new_db    
    report_str = db_name.split('-')
    report_path = report_dir + '/' + report_str[2] + '/' + report_str[1] + '/' + report_str[0]
    
    f_time = str(epoch_time(from_date_))
    t_time = str(epoch_time(to_date_))
    
    total_1 = "select count(DISTINCT L) as bu,count(D) as rq,sum(G) as tb from ( select L,D,G from main order by D ) where D between " + f_time + " and " + t_time + ";"
    total_bl = "select count(D) as bls from (select D from main where F=403 ) where D between " + f_time + " and " + t_time + ";"
    total_br = "select count(D) as brs from (select D from main where F NOT IN(503,500,407,403)) where D between " + f_time + " and " + t_time + " ;"
    u_list = "select DISTINCT users_master.A as users from (select L from main where D between " + f_time + " and " + t_time + "),users_master where L=users_master.B order by users;"
    d_list = "select A as urls from hosts_master,(select X from main where D between " + f_time + " and " + t_time + ") where X=hosts_master.B group by X order by urls;"   
    u_summary = "select users_master.A as users, rq, bt,ul,dw, bw from (select L as user, count(D) as rq, sum(E) as bt, sum(H) as ul, sum(I) as dw, sum(G) as bw from (select D,L,G,E,H,I from main order by D) where D between " + f_time + " and " + t_time + " group by L order by bw DESC),users_master where user=users_master.B;"    
    u_dn_num = "select users_master.A as users, dn from(select count(DISTINCT N) as dn, L from main where F=403  and D between " + f_time + " and " + t_time + " group by L),users_master where L=users_master.B;"
    u_al_num = "select users_master.A as users, br from(select count(DISTINCT X) as br, L from main where F NOT IN (503,500,407,403) and D between " + f_time + " and " + t_time + "  group by L),users_master where L=users_master.B;"    
    u_dm_each = "select hosts_master.A as urls, users_master.A as users, count(X) as rq, sum(G) as bw from (select X,L,G from main where F NOT IN (503,500,407,403) and  D between " + f_time + " and " + t_time + "),users_master,hosts_master where L=users_master.B and X=hosts_master.B group by X,L order by bw DESC;" 
    u_dn_fil = "select url_master.A as urls, users_master.A as users, count(N) as rq, sum(G) as bw, filter_name_master.A as fil from (select N,L,G,R from main where F=403 and D between " + f_time + " and " + t_time + " ),users_master,url_master,filter_name_master where L=users_master.B and N=url_master.B and R=filter_name_master.B group by N,L order by bw DESC;"
    total_day_bw = "select strftime('%H',datetime(D, 'unixepoch', 'localtime')) as interval, count(D) as rq, sum(G) as bw from (select D,G from main order by D) where D between " + f_time + " and " + t_time + " group by interval order by bw DESC;"       
    hour_u = "select interval, users_master.A as users, rq, bw from (select strftime('%H',datetime(D, 'unixepoch', 'localtime')) as interval, L as user, count(L) as rq,sum(G) as bw from (select D,L,G from main order by D) where D between " + f_time + " and " + t_time + " group by interval, L order by interval),users_master where user = users_master.B;"   
    total_dn_fil_1 = "select A.A as urls,B.A as fil,count(N) as rq, sum(G) as bw from url_master as A,filter_name_master as B,(select N,G,R from main where F =403 and  D between " + f_time + " and " + t_time + ") where N=A.B and R=B.B group by N,R;"
    #total_dn_fil_2 = "select hosts,group_concat(pros) from(select hosts_master.A as hosts, profiles_master.A as pros from hosts_master,profiles_master,(select A.A as pro,B.X as url from profiles as A LEFT JOIN merger.main as B  ON A.B=B.A and B.F=403 where B.D between " + f_time + " and " + t_time + ")where profiles_master.B=pro and hosts_master.B=url group by url,pro)group by hosts;"
    total_al_dm_1 = "select A as urls,count(X) as rq, sum(G) as bw from hosts_master,(select X,G from main where F NOT IN (503,500,407,403) and  D between " + f_time + " and " + t_time + ") where X=hosts_master.B group by X;"
   #total_al_dm_2 = "select hosts,group_concat(pros) from(select hosts_master.A as hosts, profiles_master.A as pros from hosts_master,profiles_master,(select A.A as pro,B.X as url from profiles as A LEFT JOIN merger.main as B  ON A.B=B.A and B.F NOT IN (503,500,407,403) where B.D between " + f_time + " and " + t_time + ")where profiles_master.B=pro and hosts_master.B=url group by url,pro)group by hosts;"

    queries = [[total_1, "total_1"], [total_bl, "total_bl"], [total_br, "total_br"], [u_list, "u_list"], [d_list, "d_list"], [u_summary, "u_summary"], [u_dn_num, "u_dn_num"], [u_al_num, "u_al_num"], [u_dm_each, "u_dm_each"], [u_dn_fil, "u_dn_fil"], [total_day_bw, "total_day_bw"], [hour_u, "hour_u"], [total_dn_fil_1, "total_dn_fil_1"], [total_al_dm_1, "total_al_dm_1"]]
    
    threadsQ = []
    for url in queries:
        t2 = generate_report_data(report_path, url, merge_full_path)
        threadsQ.append(t2)
        t2.start()
    for t2 in threadsQ:
        t2.join()
        
class generate_report_data(threading.Thread):
    def __init__(self, path, content, attach_db):
        self.url = content[0]
        self.tag = content[1]        
        self.path = path
        self.attach = attach_db
        threading.Thread.__init__(self)

    def run(self):
        start_s = time.time()
        thread_id = current_thread().name

        conn = sqlite3.connect(db_resource)

        try:
            conn.row_factory = sqlite3.Row
            curs = conn.cursor()
            logger.info('[%s] Quering for [%s] : %s', thread_id, self.tag, self.url)
            curs.executescript("attach database '"+self.attach+"' as merger;")
            curs.execute(self.url)            
            recs = curs.fetchall()
            curs.execute("detach merger;")
        except sqlite3.OperationalError as er:
            logger.error('[%s] error: %s', thread_id, er)
        except sqlite3.Error as er:
            logger.error('[%s] error: %s', thread_id, er.message)
           
        conn.close()

        try:
            rows = [dict(rec) for rec in recs]            
            logger.info('[%s] Completed query [%s] processing in : %s', thread_id, self.tag, (time.time()-start_s))
            write_file(self.path + "/." + self.tag, json.dumps(rows))
            
        except UnboundLocalError as er:            
            logger.error('[%s] error: %s', thread_id, er)

    
class Merge_db_Thread(threading.Thread):
    def __init__(self, content):
        self.days_ = content[0]        
        threading.Thread.__init__(self)

    def run(self):
        start_s = time.time()
        thread_id = current_thread().name        
        merge_db(self.days_)
        logger.info('[%s] Completed [%s] days processing in : %s', thread_id, self.days_, (time.time()-start_s))

if __name__ == "__main__":    
    arguments()    
    start_generation = time.time()
    
    #attach database
    threadsQ = []
    for i in range(0, total_days):        
        t = Merge_db_Thread([i])
        threadsQ.append(t)
        t.start()

    for t in threadsQ:
        t.join()
       
    pool = multiprocessing.Pool()    
    for i in range(total_days):                 
        pool.apply_async(query_builder, args = (i, ))
    pool.close()
    pool.join()    
  
    total_queries_ = config.get('query_match', 'queries')
    total_queries = total_queries_.split(',')
  
    logger.info('%s',days_path)
    
    for url_path in days_path:
        data_container = {}
        report_container = {}
        report_path = url_path
        
        for file_name in total_queries:
            fetch_file = report_path + "/." + str(file_name)
            
            logger.info('%s',fetch_file)
            
            try:
                with open(fetch_file, 'r') as myfile:
                    data = myfile.read()
                    data_container[file_name] = data            
                    delete_file(fetch_file)
            except FileNotFoundError as er:        
                logger.error('%s', er)
        
        dashboard_ = config.get('query_match', 'dashboard')
        dashboard = dashboard_.split(',')
        
        logger.info('%s',data_container)
        
        del dashboard_
        
        ''' dashboard  '''
        total_1 = json.loads(data_container['total_1'])
        total_bl = json.loads(data_container['total_bl'])
        total_br = json.loads(data_container['total_br'])
        total_day_bw = json.loads(data_container['total_day_bw'])
        
        report_container['bw'] = total_1[0]['tb']
        report_container['rq'] = total_1[0]['rq']
        report_container['us'] = total_1[0]['bu']
        report_container['bl'] = total_bl[0]['bls']
        report_container['al'] = total_br[0]['brs']
        report_container['ha'] = total_day_bw
        
        del total_1
        del total_bl
        del total_br
        del total_day_bw
        
        write_file(report_path + "/summary.json", json.dumps(report_container)) 
        
        report_container = {}
        
        ''' userlist database '''
        
        fetch_user = report_dir + "/user.json" 
        try:
            with open(fetch_user, 'r') as myfile:
                user_list = json.loads(myfile.read())                                
            cnt = len(user_list) + 1   
        except FileNotFoundError:        
            user_list = {}
            cnt = 1
            
        userlist = json.loads(data_container['u_list'])
        
        for s in userlist:
            current_user = s['users']        
            user_identity = "u" + str(cnt)
            userlist_len = len(user_list)
            
            if current_user == '-' :
                current_user = "unknown"
            
            if userlist_len != 0:        
                rl = reverse_lookup(user_list,current_user)            
            else:
                rl = None
                
            if rl is  None:                    
                user_list[user_identity] = current_user            
                cnt = cnt + 1
            
        report_container = user_list        
        write_file(fetch_user, json.dumps(report_container))
    
        del fetch_user
        
        report_container = {}
        
        ''' url database '''
        
        fetch_user = report_dir + "/sites.json" 
        try:
            with open(fetch_user, 'r') as myfile:
                site_list = json.loads(myfile.read())                                
            cnt = len(site_list) + 1   
        except FileNotFoundError:        
            site_list = {}            
            cnt = 1
        
        new_site_list = []
        domainlist = json.loads(data_container['d_list'])
        urllist = json.loads(data_container['total_dn_fil_1'])
        
        for s in domainlist:
            current_site = s['urls']        
            site_identity = "s" + str(cnt)
            sitelist_len = len(site_list)
            
            if current_site == '-' :
                current_site = "unknown"
            
            if sitelist_len != 0:        
                rl = reverse_lookup(site_list,current_site)            
            else:
                rl = None
                
            if rl is  None:                    
                site_list[site_identity] = current_site 
                new_site_list.append(site_identity)
                cnt = cnt + 1
            
        for s in urllist:
            current_site = s['urls']        
            site_identity = "s" + str(cnt)
            sitelist_len = len(site_list)
            
            if current_site == '-' :
                current_site = "unknown"
            
            if sitelist_len != 0:        
                rl = reverse_lookup(site_list,current_site)            
            else:
                rl = None
                
            if rl is  None:                    
                site_list[site_identity] = current_site            
                new_site_list.append(site_identity)
                cnt = cnt + 1
            
        report_container = site_list        
        write_file(fetch_user, json.dumps(report_container))
        write_file(report_path + "/newsites.json", json.dumps(new_site_list)) 
    
        report_container = {}
        container_json = []
        
        ''' allowed source '''
        alloweddomain = json.loads(data_container['total_al_dm_1'])
        
        for s in alloweddomain:
            current_site = s['urls']                    
            site_id = reverse_lookup(site_list,current_site)            
            container_json.append({'url':site_id, 'bw': s["bw"], 'rq': s["rq"], 'pro': "No"})
            
        write_file(report_path + "/allowed.json", json.dumps(container_json))
        
        container_json = []
        
        del alloweddomain
        
        ''' Denied source '''
        denieddomain = json.loads(data_container['total_dn_fil_1'])
        
        for s in denieddomain:
            current_site = s['urls']                    
            site_id = reverse_lookup(site_list,current_site)               
            container_json.append({'url':site_id, 'bw': s["bw"], 'rq': s["rq"],'fn': s["fil"], 'pro': "No"})
            
        write_file(report_path + "/denied.json", json.dumps(container_json))
        
        del denieddomain
        container_json = []
        
        ''' User report details '''
        user_summary = json.loads(data_container['u_summary'])
        
        for s in user_summary:
            current_user = s['users']  
            
            if current_user == '-' :
                current_user = "unknown"
            
            user_id = reverse_lookup(user_list,current_user)
         
            try:
                report_container[user_id]
            except KeyError:
                report_container[user_id] = {}
        
            report_container[user_id]['bw'] = s['bw']
            report_container[user_id]['tm'] = s['bt']
            report_container[user_id]['rq'] = s['rq']
            report_container[user_id]['dw'] = s['dw']
            report_container[user_id]['ul'] = s['ul']
            report_container[user_id]["hour"] = []
            report_container[user_id]["allowed"] = []
            report_container[user_id]["blocked"] = []

        user_denied = json.loads(data_container['u_dn_num']) 
        
        for s in user_denied:
            current_user = s['users'] 
            if current_user == '-' :
                current_user = "unknown"
            user_id = reverse_lookup(user_list,current_user)
            
            report_container[user_id]['bl'] = s['dn']
            
        user_allowed = json.loads(data_container['u_al_num'])     
            
        for s in user_allowed:
            current_user = s['users']  
            if current_user == '-' :
                current_user = "unknown"
            user_id = reverse_lookup(user_list,current_user)
            
            report_container[user_id]['al'] = s['br']
        
        user_hour = json.loads(data_container['hour_u'])     
            
        for s in user_hour:
            current_user = s['users']  
            if current_user == '-' :
                current_user = "unknown"
                
            user_id = reverse_lookup(user_list,current_user)
            
            report_container[user_id]['hour'].append({'int':s["interval"], 'bw': s["bw"], 'rq': s["rq"]})
        
        del user_hour
        
        user_denied_sites = json.loads(data_container['u_dn_fil'])
        
        for s in user_denied_sites:
            current_user = s['users']  
            if current_user == '-' :
                current_user = "unknown"
                
            user_id = reverse_lookup(user_list,current_user)    
            
            current_url = s['urls']  
            if current_url == '-' :
                current_url = "unknown"
                
            url_id = reverse_lookup(site_list,current_url)            
            report_container[user_id]['blocked'].append({'bl':url_id, 'bw': s["bw"], 'fn': s["fil"], 'rq': s["rq"]})
        
        del user_denied_sites
        
        user_allowed_sites = json.loads(data_container['u_dm_each'])
        
        for s in user_allowed_sites:
            current_user = s['users']  
            if current_user == '-' :
                current_user = "unknown"
                
            user_id = reverse_lookup(user_list,current_user)    
            
            current_url = s['urls']  
            if current_url == '-' :
                current_url = "unknown"
                
            url_id = reverse_lookup(site_list,current_url)            
            report_container[user_id]['allowed'].append({'al':url_id, 'bw': s["bw"], 'rq': s["rq"]})
        
        del user_allowed_sites
        
        write_file(report_path + "/report.json", json.dumps(report_container))
    
  
    end_generation = time.time()
    exhicution_time = end_generation - start_generation
    logger.info("Exhicution time taken %s",exhicution_time)
    